#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Benchmarks for testing the efficiency of the ``DynamicHashRequestFingerprinter``
and the throughput of the Selenium middleware.
"""

# Python Modules
import argparse
import logging
import pickle
import pprint
import time
import warnings

from collections import defaultdict
from typing import Any, Optional, Type

# 3rd Party Modules
from scrapy.spiders import Spider
from scrapy.crawler import Crawler, CrawlerRunner
from scrapy.http import Request, Response
from scrapy.settings import Settings, iter_default_settings
from scrapy.utils.log import configure_logging
from scrapy.utils.request import RequestFingerprinter
from tabulate import tabulate
from twisted.internet import defer, reactor

# Project Modules
from scrachy.http_ import SeleniumRequest
from scrachy.settings.defaults import selenium as selenium_defaults
from scrachy.settings.defaults.selenium import WebDriverName
from scrachy.utils.request import DynamicHashRequestFingerprinter

logging.getLogger('selenium').setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)
log = logging.getLogger('benchmark')


class BenchmarkSpider(Spider):
    name = "benchmark"
    start_urls = [
        'https://books.toscrape.com/'
    ]

    def __init__(self, request_class: Type[Request | SeleniumRequest], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.request_class = request_class

    def parse(self, response: Response, **kwargs: Any) -> Any:
        links = response.css('a::attr(href)').getall()

        for link in links:
            url = response.urljoin(link)

            if 'books.toscrape.com' in url:
                yield self.request_class(
                    url=url,
                    callback=self.parse
                )

def get_settings(
        concurrent_requests: int = 4,
        driver_name: WebDriverName = 'Chrome',
        driver_options: Optional[list[str]] = None,
        driver_extensions: Optional[list[str]] = None,
        middleware: Optional[str] = None,
        fingerprinter_class: str = 'scrapy.utils.request.RequestFingerprinter',
        implementation: str = '2.7',
        hasher: str = 'hashlib.sha1'


) -> Settings:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)

    settings['LOG_LEVEL'] = logging.INFO
    settings['DOWNLOAD_DELAY'] = 0
    settings['REQUEST_FINGERPRINTER_CLASS'] = fingerprinter_class
    settings['REQUEST_FINGERPRINTER_IMPLEMENTATION'] = implementation
    settings['CONCURRENT_REQUESTS'] = concurrent_requests
    settings['SCRACHY_FINGERPRINTER_HASHER_CLASS'] = hasher
    settings['SCRACHY_SELENIUM_WEB_DRIVER'] = driver_name
    settings['SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS'] = driver_options or ['--headless=new']
    settings['SCRACHY_SELENIUM_WEB_DRIVER_EXTENSIONS'] = driver_extensions or []

    if middleware == 'Selenium':
        settings['DOWNLOADER_MIDDLEWARES'] = {
            'scrachy.middleware.selenium.SeleniumMiddleware': 800
        }
    elif middleware == 'AsyncSelenium':
        settings['DOWNLOADER_MIDDLEWARES'] = {
            'scrachy.middleware.selenium.AsyncSeleniumMiddleware': 800
        }

    return settings


def save_requests(args: argparse.Namespace):
    log.info(f"Reading urls from '{args.url_file}'")
    with open(args.url_file, 'r') as fh:
        requests = [Request(url=url) for url in fh]

    log.info(f"Writing requests to '{args.request_file}'")
    with open(args.request_file, 'wb') as fh:
        pickle.dump(requests, fh)
    log.info(f"Done writing requests")


def make_fingerprinter(settings: Settings) -> RequestFingerprinter | DynamicHashRequestFingerprinter:
    if settings.get('REQUEST_FINGERPRINTER_CLASS').startswith('scrapy'):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            crawler = Crawler(BenchmarkSpider, settings=settings)
            return RequestFingerprinter(crawler)

    return DynamicHashRequestFingerprinter(settings)


def run_fingerprinter(settings: Settings, requests: list[Request]) -> tuple[str, str, str, dict[str, list[str]], float]:
    fingerprinter = make_fingerprinter(settings)

    fingerprints = defaultdict(list)

    total_time = 0
    for i, request in enumerate(requests):
        start = time.time()
        fingerprint = fingerprinter.fingerprint(request)
        total_time += time.time() - start

        fingerprints[fingerprint].append(request.url)

    collisions: dict[str, list[str]] = {}
    for fingerprint, urls in fingerprints.items():
        if len(urls) > 1:
            collisions[fingerprint.hex()] = urls

    return (
        fingerprinter.__class__.__name__,
        settings.get('REQUEST_FINGERPRINTER_IMPLEMENTATION'),
        settings.get('SCRACHY_FINGERPRINTER_HASHER_CLASS'),
        collisions,
        total_time
    )


def fingerprinter(args: argparse.Namespace):
    """
    5 million URL only requests using an AMD 3900X.

    ============= ============== ========= ========== ========
    Fingerprinter Implementation Hasher    Collisions Time (s)
    ============= ============== ========= ========== ========
    Scrapy        2.6            sha1      0          238.0
    Scrapy        2.6            sha1      0          231.8
    Scrachy         -            sha1      0          204.4
    Scrachy         -            xxh32     0            4.3
    Scrachy         -            xxh64     0            4.0
    Scrachy         -            xxh3_64   0           12.6
    Scrachy         -            xxh3_128  0            4.1
    Scrachy         -            spooky32  0           12.7
    Scrachy         -            spooky64  0            4.3
    Scrachy         -            spooky128 0           13.4
    ============= ============== ========= ========== ========

    :param args:
    :return:
    """
    if args.url_file and args.request_file:
        save_requests(args)
        return

    log.info(f"Start reading requests")
    if args.url_file:
        with open(args.url_file, 'r') as fh:
            requests: list[Request] = [Request(url=url) for url in fh]
    elif args.request_file:
        with open(args.request_file, 'rb') as fh:
            requests: list[Request] = pickle.load(fh)
    log.info(f"Done reading requests")

    results: list[tuple] = []
    collisions: dict[tuple[str, str, str], dict[str, list[str]]] = {}
    scrapy_fingerprinter = 'scrapy.utils.request.RequestFingerprinter'
    scrachy_fingerprinter = 'scrachy.utils.request.DynamicHashRequestFingerprinter'

    experiments = [
        (scrapy_fingerprinter, '2.6', 'hashlib.sha1'),
        (scrapy_fingerprinter, '2.7', 'hashlib.sha1'),
        (scrachy_fingerprinter, '2.7s', 'hashlib.sha1'),
        (scrachy_fingerprinter, '2.7s', 'xxhash.xxh32'),
        (scrachy_fingerprinter, '2.7s', 'xxhash.xxh64'),
        (scrachy_fingerprinter, '2.7s', 'xxhash.xxh3_64'),
        (scrachy_fingerprinter, '2.7s', 'xxhash.xxh3_128'),
        (scrachy_fingerprinter, '2.7s', 'spooky_hash.hash32'),
        (scrachy_fingerprinter, '2.7s', 'spooky_hash.hash64'),
        (scrachy_fingerprinter, '2.7s', 'spooky_hash.hash128_long'),
    ]

    for cls, impl, hasher in experiments:
        log.info(f"Starting experiment: {cls} - {impl} - {hasher}")
        tmp_result = run_fingerprinter(
            get_settings(fingerprinter_class=cls, implementation=impl, hasher=hasher),
            requests
        )
        collisions[tmp_result[:3]] = tmp_result[3]
        results.append((tmp_result[:3]) + (len(tmp_result[3]),) + (tmp_result[-1],))

    print(tabulate(results, headers=['Fingerprinter Class', 'Implementation', 'Hasher', 'Collisions', 'Time (s)']))
    print()

    print("Collisions:")
    pprint.pprint(collisions)


@defer.inlineCallbacks
def selenium(args: argparse.Namespace):
    """
    Compare the performance of the standard scrapy downloader against the
    :class:`SeleniumMiddleware` and :class:`AsyncSeleniumMiddleware`.

    The ``https://books.toscrape.com/`` site is crawled 3 times using each
    method using the number of concurrent requests specified on the command
    line.

    Example results are given below. The absolute numbers will depend on
    several factors (e.g., your internet connection and responsiveness of the
    server), but it should give a rough idea of the relative throughput for
    each method.

    The Scrapy downloader is by far the fastest way to download pages.
    The ``SeleniumMiddleware`` is bottle-necked by a single instance
    of a WebDriver and does not benefit at all from increased concurrency.
    The ``AsyncSeleniumMiddleware`` is still considerably slower than
    the default downloader, but can benefit from concurrent downloads.

    =============== =========== ======== ===========
    Method          Concurrency Requests Elapsed (s)
    =============== =========== ======== ===========
    Scrapy          2           1195     54
    Selenium        2           1195     233
    Async Selenium  2           1195     205
    Scrapy          4           1195     38
    Selenium        4           1195     233
    Async Selenium  4           1195     164
    Scrapy          8           1195     28
    Selenium        8           1195     239
    Async Selenium  8           1195     111
    Scrapy          16          1195     27
    Selenium        16          1195     235
    Async Selenium  16          1195     85
    ================================================

    :param args:
    :return:
    """
    log.info('Starting reactor')

    rows = []

    concurrent_requests = args.concurrent_requests

    log.info('Starting vanilla scrapy crawl')
    crawler, runner = setup_runner(
        'Scrapy',
        get_settings(concurrent_requests=concurrent_requests)
    )
    yield runner.crawl(crawler, request_class=Request)
    update_selenium_results('Scrapy', crawler, rows)

    log.info('Starting Selenium crawl')
    crawler, runner = setup_runner(
        'Selenium',
        get_settings(middleware='Selenium', concurrent_requests=concurrent_requests)
    )
    yield runner.crawl(crawler, request_class=SeleniumRequest)
    update_selenium_results('Selenium', crawler, rows)

    log.info('Starting AsyncSelenium crawl')
    crawler, runner = setup_runner(
        'AsyncSelenium',
        get_settings(middleware='AsyncSelenium', concurrent_requests=concurrent_requests)
    )
    yield runner.crawl(crawler, request_class=SeleniumRequest)
    update_selenium_results('Async Selenium', crawler, rows)

    print(tabulate(rows, headers=['Method', 'Concurrency', 'Requests', 'Elapsed (s)']))

    reactor.stop()  # noqa


def setup_runner(name: str, settings: Settings) -> tuple[Crawler, CrawlerRunner]:
    runner = CrawlerRunner(settings=settings)
    crawler = runner.create_crawler(BenchmarkSpider)

    return crawler, runner


def update_selenium_results(name: str, crawler: Crawler, rows: list[list[Any]]):
    stats = crawler.stats
    rows.append([
        name,
        crawler.settings.getint('CONCURRENT_REQUESTS'),
        stats.get_value('downloader/request_count'),
        stats.get_value('elapsed_time_seconds')]
    )


if __name__ == '__main__':
    parser = argparse.ArgumentParser(add_help=True)
    subparsers = parser.add_subparsers(dest='subparser')

    selenium_parser = subparsers.add_parser(
        'selenium',
        help=(
            "Run a simple spider over https://books.toscrape.com/ and compare "
            "the processing speed of the default scrapy downloader versus the "
            "two selenium middleware downloaders."
        )
    )
    selenium_parser.add_argument(
        '-c', '--concurrent-requests',
        type=int,
        default=4,
        help="The number of current requests."
    )
    selenium_parser.set_defaults(func=selenium)

    finger_parser = subparsers.add_parser(
        'fingerprinter',
        help=(
            "If either -u or -r is set exclusively, then this will benchmark "
            "the scrapy fingerprinter against the "
            "DynamicHashRequestFingerprinter using several different hash "
            "algorithms. xxhash and spooky hash must be installed. If both "
            "-u and -r are specified then the urls from -u will be used to "
            "create Responses that will be pickled to the file specified by "
            "-u. If the number of urls is large it is recommended to pickle "
            "them first."
        )
    )
    finger_parser.add_argument(
        '-u', '--url-file',
        help="A file containing a newline delimited list of urls."
    )
    finger_parser.add_argument(
        '-r', '--request-file',
        help="A file containing a list of pickled Requests."
    )
    finger_parser.set_defaults(func=fingerprinter)

    configure_logging(get_settings())

    opts = parser.parse_args()
    opts.func(opts)

    if opts.subparser == 'selenium':
        reactor.run()  # noqa



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Custom SqlAlchemy types.
"""

# Python Modules
import datetime

from typing import Any, Optional

# 3rd Party Modules
from sqlalchemy import Dialect, JSON, TypeDecorator
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import TIMESTAMP

# Project Modules


LOCAL_TIMEZONE = datetime.datetime.utcnow().astimezone().tzinfo


class TimeStampTZ(TypeDecorator):
    """
    A custom timestamp type that ensures all datetimes are timezone aware
    before entering or exiting the database.
    """
    # See: https://mike.depalatis.net/blog/sqlalchemy-timestamps.html

    impl = TIMESTAMP

    # This might produce different results if the local timezone changes,
    # but that shouldn't happen while running.
    cache_ok = True

    def __init__(self, timezone: bool = False):
        super().__init__(timezone)

    @property
    def python_type(self):
        return datetime.datetime
    
    def process_literal_param(self, value: Optional[datetime.datetime], dialect):
        if value is None:
            return None

        if value.tzinfo is None:
            value = value.astimezone(LOCAL_TIMEZONE)

        return value.astimezone(datetime.timezone.utc)

    def process_bind_param(self, value: Optional[datetime.datetime], dialect):
        if value is None:
            return None

        if value.tzinfo is None:
            value = value.astimezone(LOCAL_TIMEZONE)

        return value.astimezone(datetime.timezone.utc)

    def process_result_value(self, value: Optional[datetime.datetime], dialect):
        if value is None:
            return None

        if value.tzinfo is None:
            return value.replace(tzinfo=datetime.timezone.utc)

        return value.astimezone(datetime.timezone.utc)


class ConditionalJson(TypeDecorator):
    """Uses JSONB if the dialect is postgresql otherwise uses JSON"""
    # The common base type
    impl = JSON

    def load_dialect_impl(self, dialect: Dialect) -> TypeEngine[Any]:
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(JSONB())

        if dialect.name in ('sqlite', 'mysql', 'mssql'):
            return dialect.type_descriptor(JSON())

        raise ValueError(f"The dialect must at least support the JSON data type.")



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Database utilities.
"""

# Python Modules
from typing import Any

# 3rd Party Modules
from scrapy.settings import Settings
from sqlalchemy import URL

# Project Modules


def construct_url(settings: Settings) -> URL:
    def get(name: str) -> Any:
        return settings.get(f"SCRACHY_DB_{name}")

    driver = get('DRIVER')
    driver = f'+{driver}' if driver else ''
    drivername = f'{get("DIALECT")}{driver}'

    return URL.create(
        drivername=drivername,
        username=get('USERNAME'),
        password=get('PASSWORD'),
        host=get('HOST'),
        port=get('PORT'),
        database=get('DATABASE')
    )



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The basic data types and classes required to define the SqlAlchemy models.
"""

from __future__ import annotations

# Python Modules
import logging

from typing import Annotated

# 3rd Party Modules
from sqlalchemy import MetaData, BigInteger, LargeBinary, SmallInteger
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass, declared_attr

# Project Modules
from scrachy.settings import PROJECT_SETTINGS
from scrachy.utils.sqltypes import ConditionalJson, TimeStampTZ
from scrachy.utils.strings import camel_to_snake

bigint = Annotated[BigInteger, 64]
binary = Annotated[LargeBinary, None]
smallint = Annotated[int, 16]
timestamp = Annotated[TimeStampTZ, None]
conditional_json = Annotated[ConditionalJson, None]

log = logging.getLogger(__name__)


schema = PROJECT_SETTINGS.get('SCRACHY_DB_SCHEMA')
schema_prefix = f"{schema}." if schema else ""


class Base(DeclarativeBase, MappedAsDataclass, kw_only=True, eq=True, unsafe_hash=False):
    metadata = MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_%(constraint_name)s",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        },
        schema=schema
    )

    type_annotation_map = {
        bigint: BigInteger,
        binary: LargeBinary,
        conditional_json: ConditionalJson,
        smallint: SmallInteger,
        timestamp: TimeStampTZ(timezone=True),
    }

    # noinspection PyMethodParameters
    @declared_attr
    def __tablename__(cls):
        return camel_to_snake(cls.__name__)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
# Loosely based on: https://hackernoon.com/building-a-to-do-list-app-with-python-data-access-layer-with-sqlalchemy

"""
The Data Access Layer.
"""

# Python Modules
import abc
import datetime
import logging

from typing import Generic, Iterable, Optional, Sequence, Type, TypeVar

# 3rd Party Modules
from sqlalchemy import insert, select
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.orm import Session, load_only, selectinload

# Project Modules
from scrachy.db.base import Base
from scrachy.db.models import Response, ScrapeHistory
from scrachy.settings.defaults.storage import RetrievalMethod

BaseT = TypeVar('BaseT', bound=Base)


log = logging.getLogger(__name__)


class BaseRepository(abc.ABC, Generic[BaseT]):
    def __init__(self, model: Type[BaseT], session: Session):
        self.model = model
        self.session = session
        self.dialect = session.bind.dialect.name
        self.upsert_fn = self._get_upsert_fn()

    def find_all(self) -> Sequence[BaseT]:
        stmt = select(self.model)

        return self.session.scalars(stmt).all()

    def insert(self, obj: BaseT):
        self.session.add(obj)

    def insert_all(self, objs: Iterable[BaseT]):
        self.session.add_all(objs)

    def _get_upsert_fn(self):
        if self.dialect == 'sqlite':
            return sqlite_insert

        if self.dialect == 'postgresql':
            return pg_insert

        return insert


class ResponseRepository(BaseRepository[Response]):
    def __init__(self, session: Session):
        super().__init__(Response, session=session)

    def find_timestamp_by_fingerprint(self, fingerprint: bytes) -> Optional[datetime.datetime]:
        stmt = select(
            Response.scrape_timestamp
        ).where(
            Response.fingerprint == fingerprint
        )

        return self.session.scalars(stmt).first()

    def find_by_fingerprint(self, fingerprint: bytes, retrieval_method: RetrievalMethod = 'full') -> Optional[Response]:
        if retrieval_method == 'minimal':
            return self._find_minimal(fingerprint)

        if retrieval_method == 'standard':
            return self._find_standard(fingerprint)

        if retrieval_method == 'full':
            return self._find_full(fingerprint)

        raise ValueError(f"Unknown retrieval method: {retrieval_method}")

    def upsert(self, response: Response, returning: bool = False) -> Response:
        # If the dialect is not postgresql or sqlite, we first need to
        # query for any existing items. If one exists we should perform
        # an update. Otherwise, we can use the upsert capabilities of the
        # specific dialects.
        if self.dialect in ('sqlite', 'postgresql'):
            return self._upsert_on_conflict(response, returning)

        return self._multi_query_upsert(response)

    # region Utility Methods
    def _find_minimal(self, fingerprint: bytes) -> Response:
        stmt = select(
            Response
        ).options(
            load_only(
                Response.body
            )
        ).where(
            Response.fingerprint == fingerprint
        )

        # This should be unique
        return self.session.scalars(stmt).one_or_none()

    def _find_standard(self, fingerprint: bytes) -> Response:
        stmt = (
            select(
                Response
            ).options(
                load_only(
                    Response.body,
                    Response.meta,
                    Response.headers,
                    Response.status
                )
            ).where(
                Response.fingerprint == fingerprint
            )
        )

        return self.session.scalars(stmt).one_or_none()

    def _find_full(self, fingerprint: bytes) -> Response:
        stmt = select(
            Response
        ).options(
            selectinload(Response.scrape_history)
        ).where(
            Response.fingerprint == fingerprint,
        )

        return self.session.scalars(stmt).one_or_none()

    def _upsert_on_conflict(self, response: Response, returning: bool) -> Response:
        columns = response.__table__.columns.keys()

        stmt = self.upsert_fn(
            Response
        ).values(
            **{
                c: getattr(response, c, None)
                for c in columns
                if c != 'id'
            }
        )

        update_stmt = stmt.on_conflict_do_update(
            index_elements=[Response.fingerprint],
            set_={
                c: stmt.excluded[c]
                for c in columns
                if c not in ('id', 'fingerprint')
            }
        )

        if returning:
            update_stmt = update_stmt.returning(Response)
            result = self.session.scalars(update_stmt, execution_options={'populate_existing': True})
            return result.one_or_none()

        self.session.execute(update_stmt, execution_options={'populate_existing': True})

        return response

    def _multi_query_upsert(self, response: Response) -> Response:
        existing_response = self.find_by_fingerprint(response.fingerprint)
        update_columns = [c for c in response.__table__.columns.keys() if c not in ('id', 'fingerprint')]

        if existing_response is None:
            self.session.add(response)
            existing_response = response
        else:
            for col in update_columns:
                new_value = getattr(response, col)
                setattr(existing_response, col, new_value)

        return existing_response
    # endregion Utility Methods


class ScrapeHistoryRepository(BaseRepository[ScrapeHistory]):
    def __init__(self, session: Session):
        super().__init__(ScrapeHistory, session)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Middleware for processing requests with Selenium.
"""

from __future__ import annotations

# Python Modules
import logging
import math
import os
import queue
import signal

from struct import pack, unpack
from sys import executable
from typing import Any, Optional, cast

# 3rd Party Modules
import pickle
import psutil

from scrapy import Spider
from scrapy import signals
from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest
from scrapy.http import HtmlResponse, Request
from scrapy.settings import Settings
from selenium.common import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.protocol import ProcessProtocol
from twisted.python import failure

# Project Modules
from scrachy import PROJECT_ROOT
from scrachy.http_ import SeleniumRequest
from scrachy.cli.webdriver_server import DEFAULT_BUFFER_SIZE, Message
from scrachy.settings.defaults.selenium import WebDriverName
from scrachy.utils.selenium import ShutdownRequest, initialize_driver
from scrachy.utils.selenium import process_request as process_request_helper

log = logging.getLogger(__name__)


class SeleniumMiddleware:
    """
    A downloader middleware that uses a Selenium WebDriver to download
    the content and return an ``HtmlResponse`` if the incoming ``Response``
    is an instance of :class:`~scrachy.http_.SeleniumRequest`. Otherwise,
    it returns ``None`` to let another downloader process it.
    """
    webdriver_import_base = 'selenium.webdriver'

    def __init__(self, settings: Settings, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.settings = settings
        self._driver = self._initialize_driver()

    # region Properties
    def get(self, name: str, default_value: Any = None) -> Any:
        return self.settings.get(f'SCRACHY_SELENIUM_{name}', default_value)

    @property
    def driver_name(self) -> WebDriverName:
        return self.get('WEB_DRIVER')

    @property
    def driver_options(self) -> list[str]:
        return self.get('WEB_DRIVER_OPTIONS')

    @property
    def driver_extensions(self) -> list[str]:
        return self.get('WEB_DRIVER_EXTENSIONS')

    @property
    def driver_preferences(self) -> dict[str, Any]:
        return self.get('WEB_DRIVER_PREFERENCES')

    @property
    def driver_page_load_timeout(self) -> float:
        return self.get('PAGE_LOAD_TIMEOUT')

    @property
    def driver_implicit_wait(self) -> Optional[float]:
        return self.get('IMPLICIT_WAIT')

    @property
    def driver_verify_proxy(self) -> bool:
        return self.get('VERIFY_PROXY', False)

    @property
    def driver(self) -> WebDriver:
        return self._driver

    # endregion Properties

    # region API
    @classmethod
    def from_crawler(cls, crawler: Crawler) -> SeleniumMiddleware:
        middleware = cls(crawler.settings)

        # See: https://docs.scrapy.org/en/latest/topics/signals.html
        crawler.signals.connect(middleware.spider_closed, signals.spider_closed)

        return middleware

    def process_request(
            self,
            request: Request,
            spider: Optional[Spider] = None
    ) -> Optional[HtmlResponse | Request]:
        request = cast(SeleniumRequest, request)

        try:
            return process_request_helper(self.driver, request)
        except IgnoreRequest as e:
            raise e
        except WebDriverException as e:
            import time
            request.retries += 1
            if request.retries > request.max_retries:
                log.error(
                    f"The request for {request.url} reached the maximum number of retries and "
                    f"could not be processed."
                )
                raise IgnoreRequest()

            log.warning(
                f"A WebDriverException occurred '{e}'. "
                f"Retrying ({request.retries} out of {request.max_retries})."
            )

            request.dont_filter = True

            # Try to reinitialize the driver
            log.info("Restarting the WebDriver.")
            self.spider_closed()
            self._driver = self._initialize_driver()

            # Sleep longer and longer each time we time out
            sleep_time = request.wait_timeout * request.retries

            log.debug(f"Sleeping for {sleep_time} seconds.")
            time.sleep(sleep_time)

            return request

    def spider_closed(self, spider: Optional[Spider] = None):
        """
        Close the webdriver when the spider is closed.

        :param spider:
        :return:
        """
        if self.driver is not None:
            try:
                self.driver.quit()
            except Exception as e:
                log.warning(
                    f"Failed to quit the selenium driver. Will try to kill it.",
                    exc_info=e
                )
            finally:
                import os
                import platform

                self._driver = None
                # Sometimes there are still processes left behind even when
                # `driver.quit()` is called. I have a hunch it is related to
                # scrapy's use of twisted, but am not sure. For Chrome and
                # Firefox we can use some clues and hacks to determine the
                # parent WebDriver process and kill it manually, which should
                # kill all the other left over processes.
                if platform.system() == "Windows":
                    return

                # Although I wound up doing something different the idea was
                # inspired by the following post: https://stackoverflow.com/a/38509696
                for process in psutil.process_iter():
                    if process.status() == psutil.STATUS_ZOMBIE:
                        log.warning(
                            f"Found and terminating a zombie process with pid: {process.pid}"
                        )
                        process.terminate()
                        continue

                    cmdline = " ".join(process.cmdline()).lower()
                    if "chrom" in cmdline and  "--class=selenium" in cmdline:
                        # You can add a custom WM_CLASS to chrome using the --class flag.
                        # If the user uses the class name 'selenium' we can detect the root
                        # process and kill it here.
                        log.warning(f"Force terminating Chrome webdriver with pid: {process.pid}")
                        os.kill(process.pid, signal.SIGKILL)
                    elif "firefox" in cmdline and  "--marionette" in cmdline :
                        # The root Firefox WebDriver appears to have the --marionette flag
                        # associated with it. If it is still running kill it.
                        log.warning(f"Force terminating Firefox webdriver with pid: {process.pid}")
                        os.kill(process.pid, signal.SIGKILL)
                    # os.kill(parent_id, signal.SIGKILL)

    # endregion API
    def _initialize_driver(self) -> WebDriver:
        return initialize_driver(
            self.driver_name,
            self.driver_options,
            self.driver_extensions,
            self.driver_preferences,
            self.driver_page_load_timeout,
            self.driver_implicit_wait,
            self.driver_verify_proxy
        )


class AsyncSeleniumMiddleware:
    """
        A downloader middleware that creates a pool of Selenium WebDrivers
        and sends any incoming
        :class:`SeleniumRequests <~scrachy.http_.SeleniumRequest>` to an
        available driver to be processed.
        """
    def __init__(self, settings: Settings, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.settings = settings

        concurrent_requests: int = settings.getint('CONCURRENT_REQUESTS')
        log_file: str = settings.get('SCRACHY_SELENIUM_LOG_FILE')

        # Create a pool of drivers to increase the throughput. Since there
        # isn't actually any parallelism involved I don't think I have to
        # be all that careful with synchronization (e.g., locks).
        self.drivers = queue.Queue(maxsize=concurrent_requests)
        for driver in [WebDriverProtocol(i) for i in range(concurrent_requests)]:
            self.drivers.put(driver)

            args = [executable, '-m', 'scrachy.cli.webdriver_server']
            args += ['-d', self.driver_name]
            args += [f'-o "{o}"' for o in self.driver_options]
            args += [f'-e "{e}"' for e in self.driver_extensions]

            if log_file:
                args += [f'-f "{log_file}"']

            # noinspection PyUnresolvedReferences
            reactor.spawnProcess(
                driver,
                executable,
                args,
                path=PROJECT_ROOT,
                env=os.environ,
            )

    # region Properties
    def get(self, name: str) -> Any:
        return self.settings.get(f'SCRACHY_SELENIUM_{name}')

    @property
    def driver_name(self) -> WebDriverName:
        return self.get('WEB_DRIVER')

    @property
    def driver_options(self) -> list[str]:
        return self.get('WEB_DRIVER_OPTIONS')

    @property
    def driver_extensions(self) -> list[str]:
        return self.get('WEB_DRIVER_EXTENSIONS')
    # endregion Properties

    # region API
    @classmethod
    def from_crawler(cls, crawler: Crawler) -> AsyncSeleniumMiddleware:
        middleware = cls(crawler.settings)

        # See: https://docs.scrapy.org/en/latest/topics/signals.html
        crawler.signals.connect(middleware.spider_closed, signals.spider_closed)

        return middleware

    def process_request(
            self,
            request: Request,
            spider: Optional[Spider] = None
    ) -> Optional[Deferred[HtmlResponse | Request]]:
        if not isinstance(request, SeleniumRequest):
            # Let some other downloader handle this request
            return None

        driver = self.drivers.get()

        d = driver.process_request(request)

        def enqueue_driver(r: HtmlResponse | Request) -> HtmlResponse | Request:
            self.drivers.put(driver)
            return r

        d.addCallback(enqueue_driver)

        return d

    def spider_closed(self, spider: Optional[Spider] = None):
        # Closing stdin should shut down the server
        while not self.drivers.empty():
            driver = self.drivers.get(block=False)
            driver.shutdown()

        # Uncommenting the following lines will allow any final messages
        # sent to stderr from the server just before exiting.
        # import time
        # from twisted.internet.threads import deferToThread
        # yield deferToThread(lambda: time.sleep(0.5))
    # endregion API


class WebDriverProtocol(ProcessProtocol):
    # The number of bytes in the response message.
    response_header_size = 4

    def __init__(self, id_: int, process_buffer_size: int = DEFAULT_BUFFER_SIZE):
        # An identifier for this process
        self.id = id_

        # The size of the read buffer on the spawned process. We need to send
        # at lest this many bytes in order for the server's read buffer
        # to flush. Otherwise, the server will hang until it gets more data.
        self.process_buffer_size = process_buffer_size

        # Buffer to accumulate incoming messages
        self.buffer = b''

        # The deferred object we will eventually return
        self.deferred_response: Optional[Deferred[HtmlResponse]] = None

        # This gets set once the shutdown message is sent and will be used
        # to prevent any further communication with the protocol.
        self.is_shutdown = False

    # region Interface Methods
    def connectionMade(self):
        log.debug(f"Connection made to: {self.id} with pid: {self.transport.pid}")

    def outReceived(self, data: bytes):
        self.buffer += data
        self._extract_message()

    def errReceived(self, data: bytes):
        log.error(f"Driver process error: {data.decode()}")

    def inConnectionLost(self):
        log.debug(f"Lost stdin")

    def outConnectionLost(self):
        log.debug(f"Lost stdout")

    def errConnectionLost(self):
        log.debug(f"Lost stderr")

    def processExited(self, reason: failure.Failure):
        log.info(f"Child process exited with exit code: {reason.value.exitCode}")

    def processEnded(self, reason: failure.Failure):
        log.info(f"Child process ended: {reason.value.exitCode}")
    # endregion Interface Methods

    def process_request(self, request: SeleniumRequest) -> Deferred[HtmlResponse]:
        if self.is_shutdown:
            raise ValueError("You cannot process requests after the server has been shut down.")

        # The original request has references to all sorts of unnecessary
        # and impossible to pickle objects. Just send over what we need.
        self._send_message(
            SeleniumRequest(
                url=request.url,
                wait_timeout=request.wait_timeout,
                wait_until=request.wait_until,
                screenshot=request.screenshot,
                script_executor=request.script_executor
            )
        )

        # We'll store the response here when it is ready.
        self.deferred_response = Deferred()

        return self.deferred_response

    def shutdown(self):
        self._send_message(ShutdownRequest())

        self.transport.closeStdin()
        self.is_shutdown = True

    def _send_message(self, message: Message):
        message_data = pickle.dumps(message)

        # The number of bytes to encode the pickled data
        data_length = len(message_data)

        # The total number of bytes sent in the message (including the header
        # and padding)
        msg_length = self._get_message_length(data_length + 8)

        # The number of bytes to pad the message by. The sum of the header,
        # message, and padding should be an exact multiple of the process
        # buffer size. This is the difference between the total message length
        # and the data length and excluding the header.
        pad_length = (msg_length - data_length) - 8

        data_field = pack('!I', data_length)
        msg_field = pack('!I', msg_length)

        self.transport.writeSequence([data_field, msg_field, message_data, b' ' * pad_length])  # noqa

    def _get_message_length(self, request_length: int) -> int:
        return self.process_buffer_size * math.ceil(request_length / self.process_buffer_size)

    def _extract_message(self):
        while len(self.buffer) >= self.response_header_size:
            msg_length = unpack('!I', self.buffer[:4])[0]
            if len(self.buffer) >= msg_length + 4:
                # Get the data from the buffer
                data = self.buffer[4:4+msg_length]

                # Remove the processed data from the buffer
                self.buffer = self.buffer[4+msg_length:]

                # Try to decode the message
                try:
                    obj = pickle.loads(data)
                except pickle.PickleError as e:
                    if self.deferred_response is not None:
                        self.deferred_response.errback(e)
                    else:
                        log.error(f"There was a pickle error but the deferred response was not ready.")
                    continue

                if self.deferred_response is None:
                    log.error(f"Deferred response is not ready!")
                    continue

                if not isinstance(obj, HtmlResponse):
                    log.error(f"The message was not an HtmlResponse. Got '{type(obj)}' instead.")
                    self.deferred_response.errback(obj)
                    continue

                self.deferred_response.callback(obj)
            else:
                break  # The message is not complete



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Content extraction using `Beautiful Soup <https://www.crummy.com/software/BeautifulSoup/>`__.
"""

# Python Modules
import re

# 3rd Party Modules
import bs4
from scrapy.settings import Settings

# Project Modules
from scrachy.content import BaseContentExtractor


DOM_BLACKLIST: set[str] = {'[document]', 'noscript', 'header', 'html', 'meta', 'head', 'input'}


class BeautifulSoupExtractor(BaseContentExtractor):
    def __init__(self, settings: Settings):
        """
        A :class:`ContentExtractor` that uses
        `Beautiful Soup <https://www.crummy.com/software/BeautifulSoup/bs4/doc/>`_
        to process the HTML.

        The ``SCRACHY_CONTENT_BS4_PARSER`` setting must be set to a valid
        parser name.

        :param settings: The Scrapy ``Settings``.
        """
        super().__init__(settings)

        self.parser_name = settings.get('SCRACHY_CONTENT_BS4_PARSER')

    def get_content(self, html: str) -> str:
        """
        Extracts the textual content from the html using a simple algorithm
        described
        `here <https://matix.io/extract-text-from-webpage-using-beautifulsoup-and-python/>`_.
        In short, it ignores blocks that are unlikely to contain meaningful
        content, e.g., script blocks, and then strips the tags from the remaining
        document.

        :param html: The html content as text.
        :return: Return the extracted text.

        :param html:
        :return:
        """
        dom = bs4.BeautifulSoup(html, self.parser_name)

        # Remove script and style nodes from the DOM
        for node in dom(['script', 'style']):
            node.extract()

        # Find the remaining text nodes
        text_nodes = dom.find_all(text=True)

        # Only include the text from the nodes that aren't blacklisted
        valid_nodes = [t.strip() for t in text_nodes if t.parent.name not in DOM_BLACKLIST]

        # Normalize spaces
        valid_nodes = [re.sub(r'\s+', ' ', t) for t in valid_nodes]

        # Remove blank lines
        valid_nodes = [t for t in valid_nodes if t]

        return '\n'.join([t for t in valid_nodes])




#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Utilities to initialize and work with the SqlAlchemy engine.
"""

# Python Modules
import logging

from contextlib import contextmanager
from typing import Optional

# 3rd Party Modules
from scrapy.settings import Settings
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.ddl import CreateSchema

# Project Modules
from scrachy.db.base import Base
from scrachy.utils.db import construct_url


log = logging.getLogger(__name__)


# Singleton engine for the project. However, it is the responsibility
# of the AlchemyCacheStorage to initialize it on construction.
engine: Optional[Engine] = None
session_factory: Optional[sessionmaker] = None


def initialize_engine(settings: Settings):
    global engine
    global session_factory

    if engine is not None:
        return  # The engine is already setup

    schema = settings.get('SCRACHY_DB_SCHEMA')
    connect_args = settings.get('SCRACHY_DB_CONNECT_ARGS')

    # Create the engine
    execution_options = {"schema_translate_map": {None: schema}} if schema else None
    engine = create_engine(
        construct_url(settings),
        connect_args=connect_args,
        execution_options=execution_options
    )

    # Create the schema if necessary
    if schema is not None:
        with engine.connect() as connection:
            connection.execute(CreateSchema(schema, if_not_exists=True))
            connection.commit()

    # Create the tables if necessary
    Base.metadata.create_all(engine)

    # Create a session factory
    session_factory = sessionmaker(
        bind=engine,
        expire_on_commit=False
    )

    return engine


def reset_engine():
    global engine
    global session_factory

    if engine is not None:
        engine.dispose()

    if session_factory is not None:
        session_factory.close_all()

    engine = None
    session_factory = None


@contextmanager
def session_scope():
    if session_factory is None:
        raise ValueError("You must initialize the engine first.")

    session = session_factory()

    # noinspection PyBroadException
    try:
        yield session
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Content extraction using `BoilerPy3 <https://pypi.org/project/boilerpy3/>`__.
"""

# Python Modules
import re

# 3rd Party Modules
from scrapy.settings import Settings
from scrapy.utils.misc import load_object

# Project Modules
from scrachy.content import BaseContentExtractor


class BoilerpipeExtractor(BaseContentExtractor):
    def __init__(self, settings: Settings):
        """
        A :class:`ContentExtractor` that uses
        `BoilerPy3 <https://pypi.org/project/boilerpy3/>`__ to process the
        HTML.

        The ``SCRACHY_BOILERPY_EXTRACTOR`` must be set to a valid extractor.

        :param settings: The settings to use for initialization.
        """
        super().__init__(settings)

        self.extractor = load_object(settings.get('SCRACHY_BOILERPY_EXTRACTOR'))()

    def get_content(self, html: str) -> str:
        content = self.extractor.get_content(html)

        return self.cleanup(content)

    @staticmethod
    def cleanup(text: str) -> str:
        """
        This applies a few simple rules to clean up the text extracted from an
        html document.

            1. Split on line breaks.
            2. Strip whitespace from the beginning and ending of each line.
            3. Replace all continuous sequences of whitespace characters with a
               single space.
            4. Remove any empty lines.

        :param text: The content to clean up.
        :return: The text after applying these rules.
        """
        lines = text.splitlines()
        lines = [s.strip() for s in lines]
        lines = [re.sub(r'\s+', ' ', s) for s in lines]
        lines = [s for s in lines if s]

        return '\n'.join(lines)



##############################################################################
#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

##############################################################################
#
#   In addition to the terms listed above you must also follow the terms
#   set by the 3-Clause BSD license. See the BSD_LICENSE.md file or online
#   at <https://opensource.org/licenses/BSD-3-Clause>.
#
##############################################################################

"""
Classes for managing Http Cache Storage.
"""

# Python Modules
import datetime
import logging
import re

from typing import Optional, Any, cast

# 3rd Party Modules
from scrapy.http import Headers, Request, Response, TextResponse
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.settings import Settings
from scrapy.utils.misc import load_object
from scrapy.utils.python import to_bytes, to_unicode
from scrapy.utils.request import RequestFingerprinter
from sqlalchemy import Engine
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict

# Project Modules
from scrachy.addons import try_import
from scrachy.content import ContentExtractor
from scrachy.db.engine import initialize_engine, session_scope
from scrachy.db.models import Response as CachedResponse, ScrapeHistory
from scrachy.db.repositories import ResponseRepository, ScrapeHistoryRepository
from scrachy.exceptions import InvalidSettingError
from scrachy.settings import iter_default_settings
from scrachy.settings.defaults import filter as default_filter_settings
from scrachy.settings.defaults import fingerprinter as default_fingerprinter_settings
from scrachy.settings.defaults import storage as default_storage_settings
from scrachy.settings.defaults.storage import RetrievalMethod
from scrachy.utils.datetime import now_tzaware
from scrachy.utils.imports import get_import_path
from scrachy.utils.request import DynamicHashRequestFingerprinter, ExpirationManager
from scrachy.utils.settings import compile_patterns

log = logging.getLogger(__name__)



class BlacklistPolicy:
    """
    A wrapper around another cache control policy, but you can also blacklist
    urls (exclude from caching) via pattern matching using the
    :code:`SCRACHY_POLICY_EXCLUDE_URL_PATTERNS` setting. The patterns must
    either be strings that can be compiled with :meth:`re.compile` or
    :class:`re.Pattern` objects.
    """
    def __init__(self, settings: Settings):
        super().__init__()

        self.base_policy = load_object(settings.get('SCRACHY_POLICY_BASE_CLASS'))(settings)
        self.exclude_patterns = compile_patterns(settings.getlist('SCRACHY_POLICY_EXCLUDE_URL_PATTERNS'))

    def should_cache_request(self, request: Request) -> bool:
        if self.is_excluded(request.url):
            log.debug(f"Request url is excluded from the cache: {request.url}")
            return False

        return self.base_policy.should_cache_request(request)

    def should_cache_response(self, response: Response, request: Request) -> bool:
        if self.is_excluded(response.url):
            log.debug(f"Response url is excluded from the cache: {response.url}")
            return False

        return self.base_policy.should_cache_response(response, request)

    def is_cached_response_fresh(self, response: Response, request: Request) -> bool:
        if self.is_excluded(response.url):
            log.debug(
                f"Response url is excluded from the cache and should never be fresh: {response.url}"
            )
            return False

        return self.base_policy.is_cached_response_fresh(response, request)

    def is_excluded(self, url: str) -> bool:
        return any([p.match(url) for p in self.exclude_patterns])


class AlchemyCacheStorage:
    """
    This class implements a `scrapy cache storage backend
    <https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#writing-your-own-storage-backend>`_
    that uses a relational database to store the cached documents.
    """
    def __init__(self, settings: Settings):
        """
        This class implements a `scrapy cache storage backend
        <https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#writing-your-own-storage-backend>`_
        that uses a relational database to store the cached documents.

        :param settings: The Scrapy project middleware.
        """
        self._settings = settings

        self._expiration_manager = ExpirationManager(settings)
        self._content_extractor: Optional[ContentExtractor] = self._load_content_extractor(settings)

        # The following are created if necessary and set in the `open_spider` method
        # The callable used to fingerprint requests
        self._fingerprinter: Optional[RequestFingerprinter] = None

        # The SqlAlchemy engine used to persist the responses. This is
        # either passed in as an existing engine or created from the
        # settings in `open_spider`.
        # self._engine: Optional[Engine] = None

        # If we create the engine ourselves we should dispose of it when
        # the spider closes. Otherwise, let the creator deal with it.
        # self._dispose_on_close = None

        # A factory for creating sessions.
        # self._session_factory: Optional[sessionmaker] = None

    # region API
    # region Properties
    # Add properties for all valid middleware. At the expense of being verbose
    # this helps avoid typos and validate them (although currently there is
    # very little validation).
    def get(self, name: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        Get a ``Scrachy`` setting without having to prefix the name with ``SCRACHY_``.

        :return: The value of the setting or ``None`` if it is not set.
        """
        return self._settings.get(f"SCRACHY_{name}".upper(), default)

    @property
    def is_scrapy_fingerprinter(self) -> bool:
        """
        Returns ``True`` if we're using the scrapy fingerprinter.
        """
        return self.fingerprinter_import_path.startswith('scrapy')

    @property
    def is_scrachy_fingerprinter(self) -> bool:
        """
        Returns ``True`` if we're using the scrachy fingerprinter.
        """
        return self.fingerprinter_import_path.startswith('scrachy')

    @property
    def fingerprinter_import_path(self) -> str:
        """
        Get the import path to our fingerprinter.
        """
        return get_import_path(type(self._fingerprinter))

    @property
    def fingerprinter_implementation(self) -> str:
        """
        Get the implementation string of the fingerprinter in use.
        """
        return self._settings.get('REQUEST_FINGERPRINTER_IMPLEMENTATION')

    @property
    def fingerprinter_hasher_import_path(self) -> Optional[str]:
        if not self.is_scrachy_fingerprinter:
            return None

        hasher = cast(DynamicHashRequestFingerprinter, self._fingerprinter).hasher

        return get_import_path(hasher)

    @property
    def engine_connect_args(self) -> dict[str, Any]:
        return self.get('DB_CONNECT_ARGS', {})

    @property
    def dialect(self) -> str:
        """
        The dialect used for connecting to the database server as specified in
        the project middleware. A dialect is always required and should never be
        ``None``. For supported dialects and drivers see the
        `SQLAlchemy website <https://docs.sqlalchemy.org/en/13/core/engines.html#supported-databases>`_.

        :return: The dialect name.
        """
        return self.get('DB_DIALECT')

    @property
    def driver(self) -> Optional[str]:
        """
        The name of the driver to use with the database or ``None`` to use the
        default driver provided by SqlAlchemy.

        :return: The driver name.
        """
        return self.get('DB_DRIVER')

    @property
    def host(self) -> Optional[str]:
        """
        Returns the host name specified in the project middleware.

        :return: The host name.
        """
        return self.get('DB_HOST')

    @property
    def port(self) -> Optional[int]:
        """
        The port used to connect to the database as specified in the project
        middleware. For sqlite this should return ``None``. ``None`` also
        represents the default port for other database dialects. An error
        is raised if the port specified in the middleware can not be cast
        to an integer.

        :return: The port.
        :raises ValueError: If the port is not ``None`` and cannot be cast to
                an int.
        """
        port = self.get('DB_PORT')

        try:
            return port if port is None else int(port)
        except ValueError as e:
            log.error(f"If the port is set (not None) it must be an integer but got: {port}")
            raise e

    @property
    def database(self) -> Optional[str]:
        """
        The name of the database to connect to. The only time this should be
        ``None`` is when using an in memory sqlite database (which mostly
        defeats the purpose of a cache storage engine).

        :return: The name of the database.
        """
        return self.get('DB_DATABASE')

    @property
    def schema(self) -> Optional[str]:
        """
        The name of the schema tables will be stored in.

        :return: The name of the database.
        """
        return self.get('DB_SCHEMA')

    @property
    def default_encoding(self) -> str:
        return self.get('CACHE_DEFAULT_ENCODING', 'utf-8')

    @property
    def save_history(self) -> bool:
        return self.get('CACHE_SAVE_HISTORY')

    @property
    def activation_delay(self) -> float:
        return self.get('CACHE_ACTIVATION_SECS')

    @property
    def bs4_parser(self) -> str:
        """
        The parser to use for parsing HTML with BeautifulSoup.

        :return:
        """
        return self.get('CONTENT_BS4_PARSER')

    @property
    def response_retrieval_method(self) -> RetrievalMethod:
        """
        The name of the response retrieval method.

        This determines how much information to retrieve in the response.

        minimal
            This returns the minimal amount of information and should be the
            fastest because it does not require any joins. However, it will
            return null values for the response status and headers. Use this
            method of you don't need these or the more detailed information.
        standard
            This returns the standard information an :class:`scrapy.http.HtmlResponse`
            does.
        full
            This returns a :class:`scrachy.http.CachedResponse`, which contains
            all the information available for an item in the cache.

        :return: The type of response to retrieve.
        """
        return self.get('CACHE_RESPONSE_RETRIEVAL_METHOD')

    @property
    def expiration_secs(self) -> int:
        """
        The value of the scrapy HTTPCACHE_EXPIRATION_SECS setting.

        :return: The number of seconds before the cached item becomes stale.
                 Stale items will be re-downloaded and processed through the
                 normal pipeline regardless if they are in the cache or not.
        """
        return self._settings.getint('HTTPCACHE_EXPIRATION_SECS', 0)
    # endregion Properties

    # region Scrapy API
    def open_spider(self, spider: Spider, engine: Optional[Engine] = None):
        """
        Connect to the database, validate the middleware and set up the database
        tables if necessary.

        :param spider: The Scrapy spider.
        :param engine: Use this engine instead of creating a new one.
        """
        self._fingerprinter = spider.crawler.request_fingerprinter

        if self._fingerprinter is None:
            raise ValueError(f"The request fingerprinter has not been initialized")

        initialize_engine(spider.settings)

        self.validate_settings()

    def close_spider(self, spider: Optional[Spider] = None):
        """
        Dispose of the SqlAlchemy Engine.

        :param spider: The Scrapy spider
        """

        # self._engine.dispose()

    def retrieve_response(
            self,
            spider: Spider,
            request: Request
    ) -> Optional[Response]:
        """
        Retrieves an item from the cache if it exists, otherwise this returns
        ``None`` to signal downstream processes to continue retrieving the
        page normally. Depending on the value of the
        ``SCRACHY_RESPONSE_RETRIEVAL_METHOD`` setting more or less information
        may be returned in the response.

        :param spider: The Scrapy Spider requesting the data.
        :param request: The request describing what information to retrieve.
        :return: If the page is in the cache then this will return a
                 :class:`~scrapy.http.Response`, otherwise it will
                 return ``None``.
        """
        cached_response: CachedResponse = self._read_data(spider, request)

        # We didn't find anything (or the item is expired)
        if not cached_response:
            return None

        # Create a new Response from the cached items.
        response_retrieval_method = self.response_retrieval_method

        # Allways return this info
        response_kwargs = {
            'request': request,
            'url': request.url,
            'body': cached_response.body
        }

        meta_kwargs = dict()
        if response_retrieval_method == 'standard' or response_retrieval_method == 'full':
            if cached_response.meta:
                meta_kwargs |= cached_response.meta

            if cached_response.headers:
                raw_headers = headers_raw_to_dict(to_bytes(cached_response.headers))  # noqa
                response_kwargs['headers'] = Headers(raw_headers)

            response_kwargs['status'] = cached_response.status

        if response_retrieval_method == 'full':
            # This will return (almost) all the data Scrachy has about this
            # cached item.
            meta_kwargs |= {
                'scrape_timestamp': cached_response.scrape_timestamp,
                'extracted_text': cached_response.extracted_text,
                'body_length': cached_response.body_length,
                'extracted_text_length': cached_response.extracted_text_length,
                'scrape_history': cached_response.scrape_history
            }

        return self._make_scrapy_response(response_kwargs, meta_kwargs)

    def store_response(
            self,
            spider: Spider,
            request: Request,
            response: Response
    ):
        """
        Stores the response in the cache.

        :param spider: The Scrapy Spider issuing the request.
        :param request: The request describing what data is desired.
        :param response: The response to be stored in the cache as created by
               Scrapy's standard downloading process.
        """
        # if not isinstance(response, TextResponse):
        #     log.warning(f"The cache only supports TextResponses but received a '{type(response)}'")
        #     return

        # log.debug(f"response.body:\n{response.body}")
        # response = cast(TextResponse, response)

        fingerprint: bytes = self._fingerprinter.fingerprint(request)
        timestamp = now_tzaware()
        cached_response = self._make_cached_response(fingerprint, timestamp, request, response)

        # with self.session_scope() as session:
        with session_scope() as session:
            response_repo = ResponseRepository(session)
            response_repo.upsert(cached_response)

            if self.save_history:
                history_repo = ScrapeHistoryRepository(session)
                history_repo.insert(
                    ScrapeHistory(
                        fingerprint=fingerprint,
                        scrape_timestamp=timestamp,
                        body=response.text
                    )
                )
    # endregion Scrapy API

    def validate_settings(self):
        """
        This makes sure that any setting starting with the prefix ``SCRACHY``
        is known to the storage backend.

        It performs some minor validation like checking to make sure the
        port is an integer and a host name is specified unless the dialect is
        sqlite. It is still primarily up to the user to ensure the database
        connection properties are valid for the type of database being used.

        :raises InvalidSettingError: If there are:

                * unknown scrachy middleware.
                * invalid database middleware.
                * an option to a setting that is not valid.
                * the hash algorithm specified in the project middleware used to
                  create the request fingerprint is different from the one
                  already used for this cache region.
        """
        self._validate_unknown_settings()
        self._validate_supported_options()
        self._validate_database_parameters()

    @staticmethod
    def clear_cache():
        with session_scope() as session:
            session.query(Response).delete()
            session.query(Request).delete()

    def dump_cache(self) -> list[Response]:
        """
        Dump the contents of the cache. This is not recommended except for
        debugging.

        :return: A list of SQLAlchemy result objects that contains all
                 the items in the cache.
        """
        with session_scope() as session:
            return list(ResponseRepository(session).find_all())
    # endregion Extended API
    # endregion API

    # region Utility Methods
    # region Initialization
    @staticmethod
    def _load_content_extractor(settings: Settings) -> Optional[ContentExtractor]:
        extraction_obj = settings.get('SCRACHY_CONTENT_EXTRACTOR')

        if not extraction_obj:
            return None

        return load_object(extraction_obj)(settings)
    # endregion Initialization

    # region Validation
    def _validate_unknown_settings(self):
        """
        Check that there aren't any unknown scrachy middleware (e.g., typos).

        :raises InvalidSettingError:
        """
        storage_re = re.compile(f'^SCRACHY_(CACHE|DB|USE_CONTENT|CONTENT|USE_SIMHASH|SIMHASH)')
        storage_keys = set([k for k, _ in iter_default_settings(default_storage_settings)])
        finger_keys = set([k for k, _ in iter_default_settings(default_fingerprinter_settings)])
        filter_keys = set([k for k, _ in iter_default_settings(default_filter_settings)])

        valid_keys = storage_keys | finger_keys | filter_keys

        for key, value in self._settings.items():
            if storage_re.match(key) and key not in valid_keys:
                raise InvalidSettingError(f"Unknown scrachy setting: {key}")

    def _validate_supported_options(self):
        """
        Make sure the parameters for middleware that take one are valid.

        :raises InvalidSettingError:
        """
        if 'lxml' in self.bs4_parser:
            try_import('lxml', 'AlchemyCacheStorageAddon')

        if self.bs4_parser == 'html5lib':
            try_import('html5lib', 'AlchemyCacheStorageAddon')

    def _validate_database_parameters(self):
        """
        Check to make sure the database parameters are valid.

        :raises InvalidSettingError:
        """

        if self.port and not isinstance(self.port, int):
            raise InvalidSettingError(
                f"If the port is specified it must be an integer, but was: {self.port}"
            )

        if self.dialect != 'sqlite' and self.database is None:
            raise InvalidSettingError(
                "You must specify a database name for dialects except sqlite."
            )

        if self.dialect == 'sqlite' and self.driver == 'pysqlcipher':
            raise InvalidSettingError("The pysqlcipher driver is not supported.")

    # region Retrieve Response Utilities
    def _read_data(self, spider: Spider, request: Request) -> Optional[CachedResponse]:
        # In an effort to save storage space in the previous version I stored
        # the diff of the body if there was already an entry. I think this
        # was misguided, because it traded disk space for latency by having
        # to retrieve multiple objects from the database and then apply the
        # diffs to get the actual body. So, in this version I just store
        # everything no matter what. It also adds a lot of additional complexity
        # that is difficult to maintain.
        fingerprint = self._fingerprinter.fingerprint(request)

        # with self.session_scope() as session:
        with session_scope() as session:
            repo = ResponseRepository(session)
            scrape_timestamp = repo.find_timestamp_by_fingerprint(fingerprint)

            # A missing timestamp indicates the data is not in the cache.
            if not scrape_timestamp:
                log.debug(
                    f"Did not find the timestamp for '{request.url}' "
                    f"with fingerprint '{fingerprint.hex()}' in the cache."
                )
                return None

            # Don't use the cached data if it is stale.
            # Note the CachePolicy does something different. It only uses
            # info available from the page itself (e.g., the headers) and
            # doesn't know anything about how long (if at all) the data
            # has been in our cache.
            if self._expiration_manager.is_stale(request.url, scrape_timestamp):
                log.debug(
                    f"The response with url '{request.url}' "
                    f"and timestamp '{scrape_timestamp}' is stale"
                )
                return None

            cached_response = repo.find_by_fingerprint(fingerprint, self.response_retrieval_method)

            session.expunge_all()

        return cached_response
    # endregion Retrieve Response Utilities

    # region Common Utilities
    def _make_cached_response(
            self,
            fingerprint: bytes,
            timestamp: datetime.datetime,
            request: Request,
            response: Response
    ):
        cached_response = CachedResponse(
            fingerprint=fingerprint,
            scrape_timestamp=timestamp,
            url=request.url,
            body=response.body,
            status=response.status,
            body_length=len(response.body)
        )

        # For some reason this doesn't work as a constructor parameter anymore
        cached_response.request_method = request.method

        if request.body is not None:
            cached_response.request_body = request.body

        if response.request is not None:
            cached_response.meta = request.meta

        if response.headers is not None:
            cached_response.headers = to_unicode(headers_dict_to_raw(response.headers))

        if self._content_extractor is not None and isinstance(response, TextResponse):
            cached_response.extracted_text = self._content_extractor.get_content(response.text)
            cached_response.extracted_text_length = len(to_bytes(cached_response.extracted_text))

        return cached_response

    def _make_scrapy_response(
            self,
            response_kwargs: dict[str, Any],
            meta_kwargs: dict[str, Any]
    ) -> Response:
        response_cls = responsetypes.from_args(
            headers=response_kwargs.get('headers'),
            url=response_kwargs.get('url'),
            body=response_kwargs.get('body')
        )

        if issubclass(response_cls, TextResponse):
            response = response_cls(encoding=self.default_encoding, **response_kwargs)
        else:
            response = response_cls(**response_kwargs)

        if response.request is not None:
            response.meta.update(meta_kwargs)

        return response

    # endregion Common Utilities
    # endregion Utility Methods



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
An enhanced ``RequestFingerprinter``.
"""

from __future__ import annotations

# Python Modules
import datetime
import re

from typing import Iterable, Optional, Tuple, TypeVar, Union
from weakref import WeakKeyDictionary

# 3rd Party Modules
import msgspec

from cron_converter import Cron
from scrapy import Request
from scrapy.crawler import Crawler
from scrapy.settings import Settings
from scrapy.utils.misc import load_object
from scrapy.utils.python import to_bytes
from w3lib.url import canonicalize_url

# Project Modules
from scrachy.settings.defaults.storage import PatternLike, Schedulable
from scrachy.utils.datetime import now_tzaware

ExpirationType = TypeVar('ExpirationType')

DEFAULT_HASHER_CLASS = 'hashlib.sha1'
DEFAULT_SCRACHY_FINGERPRINTER_VERSION = 'scrachy_2.7'

_fingerprint_cache = WeakKeyDictionary()


class DynamicHashRequestFingerprinter:
    """
    Almost identical to the Scrapy version 2.7 algorithm, but allows you
    to configure which hash algorithm is used.
    """
    def __init__(self, settings: Optional[Settings] = None):
        super().__init__()

        self.hasher = load_object(
            settings.get('SCRACHY_REQUEST_FINGERPRINTER_HASHER_CLASS', DEFAULT_HASHER_CLASS)
        )

        self.encoder = msgspec.msgpack.Encoder()

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> DynamicHashRequestFingerprinter:
        return cls(crawler.settings)

    def fingerprint(
        self,
        request: Request,
        *,
        include_headers: Optional[Iterable[Union[bytes, str]]] = None,
        keep_fragments: bool = False,
    ) -> bytes:
        """
        Return the request fingerprint.

        The request fingerprint is a hash that uniquely identifies the resource the
        request points to. For example, take the following two urls:

        http://www.example.com/query?id=111&cat=222
        http://www.example.com/query?cat=222&id=111

        Even though those are two different URLs both point to the same resource
        and are equivalent (i.e. they should return the same response).

        Another example are cookies used to store session ids. Suppose the
        following page is only accessible to authenticated users:

        http://www.example.com/members/offers.html

        Lots of sites use a cookie to store the session id, which adds a random
        component to the HTTP Request and thus should be ignored when calculating
        the fingerprint.

        For this reason, request headers are ignored by default when calculating
        the fingerprint. If you want to include specific headers use the
        include_headers argument, which is a list of Request headers to include.

        Also, servers usually ignore fragments in urls when handling requests,
        so they are also ignored by default when calculating the fingerprint.
        If you want to include them, set the keep_fragments argument to True
        (for instance when handling requests with a headless browser).
        """
        processed_include_headers: Optional[Tuple[bytes, ...]] = None
        if include_headers:
            processed_include_headers = tuple(to_bytes(h.lower()) for h in sorted(include_headers))

        cache = _fingerprint_cache.setdefault(request, {})
        cache_key = (processed_include_headers, keep_fragments)

        if cache_key not in cache:
            # Unlike json, msgspec should be able to handle bytes
            if processed_include_headers:
                headers = {
                    k: request.headers.values()
                    for k in processed_include_headers
                    if k in request.headers
                }
            else:
                headers: dict[bytes, list[bytes]] = {}

            fingerprint_data = {
                'method': to_bytes(request.method),
                'url': to_bytes(canonicalize_url(request.url, keep_fragments=keep_fragments)),
                'body': request.body or b'',
                'headers': headers
            }
            fingerprint_msgpack = self.encoder.encode(fingerprint_data)
            cache[cache_key] = self.hasher(fingerprint_msgpack).digest()

        return cache[cache_key]


class ExpirationPattern:
    def __init__(self, patterns: list[tuple[PatternLike, ExpirationType]], default_value: ExpirationType):
        self.patterns = self._initialize_patterns(patterns)
        self.default_value = default_value
        self.cache: dict[str, ExpirationType] = dict()

    def __call__(self, url: str) -> ExpirationType:
        if url in self.cache:
            return self.cache[url]

        for pattern, value in self.patterns:
            if pattern.match(url):
                self.cache[url] = value
                return value

        self.cache[url] = self.default_value
        return self.default_value

    @staticmethod
    def _initialize_patterns(
            patterns: list[tuple[PatternLike, ExpirationType]]
    ) -> list[tuple[re.Pattern, ExpirationType]]:
        return [(re.compile(p), v) for p, v in patterns]


class ExpirationManager:
    def __init__(self, settings: Settings):
        self.activation_matcher = ExpirationPattern(
            settings.getlist('SCRACHY_CACHE_ACTIVATION_SECS_PATTERNS'),
            settings.getfloat('SCRACHY_CACHE_ACTIVATION_SECS', 0.0)
        )

        self.expiration_matcher = ExpirationPattern(
            settings.getlist('SCRACHY_CACHE_EXPIRATION_SECS_PATTERNS'),
            settings.getfloat('HTTPCACHE_EXPIRATION_SECS', 0)
        )

        patterns = [
            (k, self._initialize_schedule(v))
            for k, v in settings.getlist('SCRACHY_CACHE_EXPIRATION_SCHEDULE_PATTERNS')
        ]
        self.schedule_matcher = ExpirationPattern(
            patterns,
            self._initialize_schedule(settings.get('SCRACHY_CACHE_EXPIRATION_SCHEDULE'))
        )

    def is_stale(
            self,
            url: str,
            scrape_timestamp: datetime.datetime,
            current_timestamp: Optional[datetime.datetime] = None
    ) -> bool:
        return not self.is_fresh(url, scrape_timestamp, current_timestamp)

    def is_fresh(
            self,
            url: str,
            scrape_timestamp: datetime.datetime,
            current_timestamp: Optional[datetime.datetime] = None
    ) -> bool:
        current_timestamp = current_timestamp or now_tzaware()

        secs_in_cache = (current_timestamp - scrape_timestamp).total_seconds()

        if secs_in_cache < self.activation_matcher(url):
            # The item hasn't been in the cache long enough
            return False

        expiration_secs = self.expiration_matcher(url)
        if 0 < expiration_secs < secs_in_cache:
            # The item has been in the cache too long
            return False

        if current_timestamp > self.get_expiration_date(url, scrape_timestamp):
            # The current time is past the expiration date (regardless of
            # how long the item has been in the cache)
            return False

        return True

    def get_expiration_date(self, url: str, timestamp: datetime.datetime) -> datetime.datetime:
        cron: Optional[Cron] = self.schedule_matcher(url)

        if cron is None:
            return datetime.datetime(datetime.MAXYEAR, 12, 31).replace(tzinfo=datetime.timezone.utc)

        return cron.schedule(timestamp).next()

    @staticmethod
    def _initialize_schedule(cron: Optional[Schedulable]) -> Optional[Cron]:
        if cron is None:
            return None

        if isinstance(cron, str):
            return Cron(cron)

        if isinstance(cron, Cron):
            return cron

        raise ValueError(f"Unknown scheduler class: {type(cron)}")



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The default settings for configuring
:class:`~scrachy.middleware.httpcache.BlacklistPolicy`.
"""

# Python Modules
import re

from typing import Callable

# 3rd Party Modules

# Project Modules


SCRACHY_POLICY_BASE_CLASS: str | Callable = 'scrapy.extensions.httpcache.DummyPolicy'
"""
The base policy the :class:`~scrachy.middleware.httpcache.BlacklistPolicy` 
will wrap around. The policy can be specified as the full import path to
the class or a class object itself. Either way the class constructor must 
accept a :class:`~scrapy.settings.Settings` object as its first parameter. 
"""

SCRACHY_POLICY_EXCLUDE_URL_PATTERNS: list[str | re.Pattern] = []
"""
A list of regular expression patterns, as compilable strings or 
:class:`re.Pattern` objects. Any url matching any of these patterns will be 
excluded from being processed by the 
:class:`~scrapy.downloadermiddlewares.httpcache.HttpCacheMiddleWare`.
"""





#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Some utility classes for sending messages between the
:class:`~scrachy.middleware.selenium.AsyncSeleniumMiddleware` and the
:mod:`~scrachy.cli.webdriver_server`. It also includes the primary
functionality for processing requests with Selenium. Each Selenium middleware
is a thin wrapper around these functions.
"""

from __future__ import annotations

# Python Modules
import logging
import time

from typing import Any, Optional, Type, cast

# 3rd Party Modules
from scrapy.exceptions import IgnoreRequest
from scrapy.http import Request, HtmlResponse
from scrapy.utils.misc import load_object
from selenium import webdriver
from selenium.common import TimeoutException, WebDriverException
from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait

# Project Modules
from scrachy.http_ import SeleniumRequest
from scrachy.settings.defaults.selenium import WebDriverName


log = logging.getLogger(__name__)


webdriver_import_base = 'selenium.webdriver'


class BufferIncompleteError(Exception):
    """Raised when the amount of data received is less than expected."""
    def __init__(self, *args):
        super().__init__(*args)


class ShutdownRequest:
    """A message token indicates the webdriver server should terminate."""
    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnknownMessageType:
    """
    A message token indicates the webdriver server received an unknown message
    type.
    """
    def __init__(self, message_type: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.message_type = message_type


class UnhandledError:
    """
    A message token indicates that an unhandled exception was raised during the
    processing of a ``SeleniumRequest``.
    """
    def __init__(self, exception: Exception, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.exception = exception


def initialize_driver(
        driver_name: WebDriverName,
        options: list[str],
        extensions: list[str],
        preferences: Optional[dict[str, Any]] = None,
        page_load_timeout: Optional[float] = None,
        implicit_wait: Optional[float] = None,
        verify_proxy: bool = False
) -> WebDriver:
    """
    Construct a new webdriver instance with the given options and extensions.

    :param page_load_timeout:
    :param preferences:
    :param driver_name: The class name of the webdriver to construct.
    :param options: The list of options for the webdriver to use.
    :param extensions: The list of extensions to install.
    :param implicit_wait: The amount of time to implicitly wait.
    :param verify_proxy: If ``True`` check that the IP address of a request
           from the host machine is not the same as a request from the
           webdriver (if using a proxy).

    :return: The webdriver.
    """
    driver_name: WebDriverName = driver_name
    driver_cls: Type[WebDriver] = load_object(f'{webdriver_import_base}.{driver_name}')
    driver_options: ArgOptions = load_object(f'{webdriver_import_base}.{driver_name}Options')()

    for option in options:
        if not option.startswith("--class="):
            log.debug(f"Adding option '{option}' to the WebDriver.")
            driver_options.add_argument(option)
        else:
            log.warning(
                f"The --class option is reserved for use by Scrachy to help clean up "
                f"rogue processes that are not terminated by `driver.quit()`."
            )

    if preferences:
        log.debug(f"Setting the WebDriver preferences: {preferences}")
        driver_options.add_experimental_option("prefs", preferences)  # noqa

    if "chrom" in driver_name.lower():
        driver_options.add_argument("--class=selenium")

    # Chrome loads the extensions from the options
    if driver_name == 'Chrome' and extensions:
        driver_options = cast(ChromiumOptions, driver_options)
        for extension in extensions:
            driver_options.add_extension(extension)  # noqa

    driver = driver_cls(options=driver_options)

    # Firefox appears to load the extension directly from the driver
    if driver_name == 'Firefox' and extensions:
        for extension in extensions:
            driver = cast(webdriver.Firefox, driver)
            driver.install_addon(extension, temporary=True)

    # implicitly_wait applies to all find element calls.
    # set_page_load_timeout applies to 'get' and 'navigate' calls.
    if page_load_timeout is not None:
        driver.set_page_load_timeout(page_load_timeout)

    if implicit_wait is not None:
        driver.implicitly_wait(implicit_wait)

    if verify_proxy:
        my_ip = _get_local_ip()
        driver_ip = _get_driver_ip(driver)

        log.debug(f"local_ip: {my_ip}, proxy_ip: {driver_ip}")
        if my_ip == driver_ip:
            raise ValueError("The local IP address is the same as the driver IP address")

    return driver


def process_request(driver: WebDriver, request: Request) -> Optional[HtmlResponse | Request]:
    if not isinstance(request, SeleniumRequest):
        # Let some other downloader handle this request
        return None

    request = cast(SeleniumRequest, request)

    try:
        driver.get(request.url)
        set_cookies(driver, request)
    except TimeoutException:
        raise WebDriverException("Failed to get the request because of a page load timeout.")
        # request.retries += 1
        # if request.retries > request.max_retries:
        #     log.error(
        #         f"The request for {request.url} reached the maximum number of retries and "
        #         f"could not be processed."
        #     )
        #     raise IgnoreRequest()
        #
        #
        #
        #
        # # Don't increase the wait time on each retry
        # page_load_timeout = driver.timeouts.page_load
        # sleep_time = page_load_timeout
        # log.warning(
        #     f"The request {request} reached the page_load_timeout of {page_load_timeout}s. "
        #     f"Sleeping for {sleep_time}s and then rescheduling the request "
        #     f"({request.retries} out of {request.max_retries})."
        # )
        #
        # time.sleep(sleep_time)
        #
        # request.dont_filter = True
        # return request

    try:
        wait_for_page(driver, request)
    except TimeoutException:
        # The timeout could actually be from a page timeout (i.e., from the
        # call to `driver.get`) or from a WebDriverWait condition (i.e.,
        # a call to `wait_for_page`). It doesn't really matter, but log
        # message will be incorrect for a page timeout.
        request.retries += 1
        if request.retries > request.max_retries:
            log.error(
                f"The request for {request.url} reached the maximum number of retries and "
                f"could not be processed."
            )
            raise IgnoreRequest()

        # Sleep longer and longer each time we time out
        sleep_time = request.wait_timeout * request.retries

        log.warning(
            f"The page for request {request} did not load properly after {request.wait_timeout} "
            f"seconds. Sleeping for {sleep_time}s and then rescheduling the request "
            f"({request.retries} out of {request.max_retries})."
        )

        time.sleep(sleep_time)

        request.dont_filter = True
        return request

    take_screenshot(driver, request)

    execute_script(driver, request)

    response = make_response(driver, request)

    return response


def set_cookies(driver: WebDriver, request: SeleniumRequest):
    for cookie_name, cookie_value in request.cookies.items():
        driver.add_cookie({'name': cookie_name, 'value': cookie_value})


def wait_for_page(driver: WebDriver, request: SeleniumRequest):
    if request.wait_until:
        try:
            WebDriverWait(
                driver,
                request.wait_timeout,
                poll_frequency=request.poll_frequency
            ).until(
                request.wait_until
            )
        except TimeoutException as e:
            log.error(f"Wait condition timed out for url: '{request.url}'")
            raise e


def take_screenshot(driver: WebDriver, request: SeleniumRequest):
    if request.screenshot:
        request.meta['screenshot'] = driver.get_screenshot_as_png()


def make_response(driver: WebDriver, request: SeleniumRequest) -> HtmlResponse:
    return HtmlResponse(
        url=driver.current_url,
        body=driver.page_source,
        encoding='utf-8',
        request=request,
        flags=["selenium"]
    )


def execute_script(driver: WebDriver, request: SeleniumRequest):
    if request.script_executor is not None:
        request.meta['script_result'] = request.script_executor(driver, request)


def _get_local_ip() -> str:
    import requests

    response = requests.get('https://api.ipify.org?format=json')
    return response.json()['ip']


def _get_driver_ip(driver: WebDriver) -> str:
    import json
    driver.get('https://api.ipify.org?format=json')
    data = driver.find_element(By.CSS_SELECTOR, 'body')
    return json.loads(data.text)['ip']



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Anything related to the settings for this project.
"""

# Python Modules
from types import ModuleType
from typing import Any, Iterable, Optional, Tuple

# 3rd Party Modules
from scrapy.utils.project import get_project_settings

# Project Modules

PROJECT_SETTINGS = get_project_settings()


def iter_default_settings(settings_module: Optional[ModuleType] = None) -> Iterable[Tuple[str, Any]]:
    """
    Similar to :func:`scrapy.middleware.iter_default_settings`, but accepts
    the module import path to use.

    :param settings_module: The module where the middleware are located.
    :return: An iterator of (name, value) tuples.
    """
    if settings_module is None:
        return []

    for name in dir(settings_module):
        if name.isupper():
            yield name, getattr(settings_module, name)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The addons provided by Scrachy.
"""

from __future__ import annotations

# Python Modules
import importlib
import logging
from types import ModuleType

from typing import Optional

# 3rd Party Modules
from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings

# Project Modules
from scrachy.settings.defaults import fingerprinter, storage
from scrachy.settings.defaults import filter as cache_filter
from scrachy.utils.imports import get_import_path
from scrachy.utils.request import DEFAULT_SCRACHY_FINGERPRINTER_VERSION

log = logging.getLogger(__name__)


def try_import(module_name: str, addon_name):
    """
    Try importing a module by name and raise a
    :class:`scrapy.exceptions.NotConfigured` error if it can't be found.

    :param module_name: The full path to the module.
    :param addon_name: The name of the addon.
    """
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise NotConfigured(f"The {addon_name} addon requires the module: {module_name}")


class ScrachyAddon:
    """
    The base class for all Scrachy addons.
    """
    def __init__(self, settings_module: Optional[ModuleType] = None):
        self.settings_module = settings_module

    def update_settings(self, settings: Settings):
        settings.setmodule(self.settings_module, 'addon')


class BlacklistPolicyAddon(ScrachyAddon):
    def __int__(self):
        super().__init__()


class DynamicHashRequestFingerprinterAddon(ScrachyAddon):
    def __init__(self):
        super().__init__(fingerprinter)

    def update_settings(self, settings: Settings):
        super().update_settings(settings)

        fp = get_import_path(settings.get('REQUEST_FINGERPRINTER_CLASS'))

        if fp.startswith('scrachy'):
            settings['REQUEST_FINGERPRINTER_IMPLEMENTATION'] = DEFAULT_SCRACHY_FINGERPRINTER_VERSION


class AlchemyCacheStorageAddon(ScrachyAddon):
    def __init__(self):
        super().__init__(storage)

    def update_settings(self, settings: Settings):
        def check_import(m: str):
            try_import(m, self.__class__.__name__)

        check_import('bs4')
        check_import('w3lib')
        check_import('sqlalchemy')

        super().update_settings(settings)


class CachedResponseFilterAddon(ScrachyAddon):
    def __init__(self):
        super().__init__(cache_filter)



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with Foobar.  If not, see <https://www.gnu.org/licenses/>.

"""
Scrachy provides several middleware and utility classes to improve the
functionality of scrapy.
"""

# Python Modules
import pathlib

# 3rd Party Modules

# Project Modules

__version__ = '0.7.0'

PROJECT_ROOT = pathlib.Path(__file__).parent.parent



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The SqlAlchemy models.
"""

from __future__ import annotations

# Python Modules
from typing import Optional

# 3rd Party Modules
import sqlalchemy
import sqlalchemy.orm

from sqlalchemy import ColumnElement, String, ForeignKey, type_coerce
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, relationship

# Project Modules
from scrachy.db.base import Base, binary, conditional_json, smallint, timestamp


# Primary tables
class Response(Base):
    """
    Stores the minimum amount of information to recreate the html of a
    page given its request fingerprint. The cache only stores the most
    recent version of the page, but the :class:`ScrapyHistory` table
    can be used to view

    Additional data is stored in other tables, but can be accessed via
    relationships.
    """
    # A hash of the page's uniquely identifying request information.
    fingerprint: Mapped[binary] = mapped_column(primary_key=True)

    # The most recent timestamp when the page was scraped.
    scrape_timestamp: Mapped[timestamp] = mapped_column(default=None)

    # This is redundant with the request, but can prevent a join when
    # we might want to search or filter responses by it.
    url: Mapped[str] = mapped_column(index=True)

    # The request method (True for GET. False for POST)
    method: Mapped[bool] = mapped_column(default=None)

    # The body of the request (primarily used for POST requests)
    request_body: Mapped[Optional[binary]] = mapped_column(default=None)

    # The response body
    body: Mapped[binary] = mapped_column(default=None)

    # Store the metadata associated with the response as JSON.
    meta: Mapped[Optional[conditional_json]] = mapped_column(default=None)

    # The response status code for the page
    status: Mapped[Optional[smallint]] = mapped_column(default=None)

    # The response headers
    headers: Mapped[Optional[str]] = mapped_column(default=None)

    # The text of the response stripped of HTML and possibly stripped of
    # boilerplate content.
    extracted_text: Mapped[Optional[str]] = mapped_column(default=None)

    # The number of bytes of the full downloaded response (including the HTML)
    body_length: Mapped[int]

    # The number of bytes of the response excluding the HTML.
    extracted_text_length: Mapped[Optional[int]] = mapped_column(default=None)

    # The scrape history
    scrape_history: Mapped[list[ScrapeHistory]] = relationship(
        cascade='all, delete-orphan',
        default_factory=list,
        repr=False
    )

    # I believe hybrid properties allow changing how the property behaves
    # depending on whether you are using it as an instance or as a class
    # template. For example, when operating on an instance referencing the
    # `body` property will use the regular python code in the method body.
    # But if you using it as a template for SQL operations you can define
    # a @xxx.inplace.expression method that will use the same property name
    # but emit SQL code instead.

    # So here, for example, it uses less space to store the request method
    # as a boolean (because it can only ever have two values), but it is
    # more convenient to work with these values as their original string
    # representations.
    @hybrid_property
    def request_method(self) -> Optional[str]:
        """
        Get the request method as a string instead of the stored boolean
        type.
        """
        if self.method is True:
            return 'GET'
        elif self.method is False:
            return 'POST'

        return None

    @request_method.inplace.setter
    def request__method(self, method: str) -> None:
        """
        Client side setter.

        :param method:
        :return:
        """
        self.method = method.upper() == 'GET'  # noqa

    # noinspection PyMethodParameters,PyNestedDecorators
    @request_method.inplace.expression
    @classmethod
    def request_method(cls) -> ColumnElement[Optional[str]]:
        """
        Server side setter.

        :return:
        """
        return type_coerce(
            sqlalchemy.case(
                (cls.method == True, 'GET'),
                (cls.method == False, 'POST'),
            ),
            String
        )


class ScrapeHistory(Base):
    """
    Store the history of when a page was scraped.
    """
    fingerprint: Mapped[binary] = mapped_column(ForeignKey('response.fingerprint'), primary_key=True)

    # When the page was scraped
    scrape_timestamp: Mapped[timestamp] = mapped_column(primary_key=True)

    # The response body
    body: Mapped[binary]
